--
-- Piglet manager, to support fourth-party "plugin plugins" for Adobe
-- Lightroom export plugins.
--
-- Copyright 2008 Jeffrey Friedl
--
-- http://regex.info/blog/photo-tech/piglets/
-- http://regex.info/blog/
--

local LrFunctionContext = import 'LrFunctionContext'
local LrPathUtils       = import 'LrPathUtils'
local LrFileUtils       = import 'LrFileUtils'
local LrBinding         = import 'LrBinding'
local LrDialogs         = import 'LrDialogs'
local LrErrors          = import 'LrErrors'
local LrPrefs           = import 'LrPrefs'
local LrColor           = import 'LrColor'
local LrTasks           = import 'LrTasks'
local LrView            = import 'LrView'
local LrHttp            = import 'LrHttp'

local Piglets = {
   piglet_manager_version = 3, --  2007-Jan-22
}

-- Change Log
--
-- Version 3 (Jan 22, 2007) -- Isolates errors in piglets from crashing host plugin.
--                             Added ability to set the ordering when multiple piglets are used.
--                             Added ability to clear out old piglet info once a piglet file has been removed.
--
-- Version 2 (Jan 17, 2007) -- added Piglets.allowFileFormats
--
-- Version 1 (Jan 15, 2007) -- Initial releae
--


local PIGLETS

--
-- Given two tables, append all the stuff in the second table to the first (in place)
-- Returns the first.
--
function concatenate(t1, t2)
   for key, value in pairs(t2) do
      if type(key) == "number" then
         table.insert(t1, value)
      else
         t1[key] = value
      end
   end
   return t1
end


--require 'jfriedlDebug.lua' -- defines Dump() and DebugNote()

local function runtime_error(piglet_info, piglet, action, error_message)

   if piglet then
      piglet.invalid = true
   end
   piglet_info.invalid = true
   piglet_info.invalid_reason = LOC("$$$/PigletManager/10=Runtime error in ^1: ^2", action, error_message)

   LrDialogs.message(LOC("$$$/PigletManager/11=Runtime error in ^[^1^] piglet in ^2.^nPiglet Disabled.", piglet_info.piglet_title, action),
                     error_message,
                     "critical")
   --LrErrors.throwUserError(piglet_info.invalid_reason)
end

local function read_piglet(state, piglet_info, basic_info_only)

   piglet_info.invalid = true -- we'll clear this if things seem to be fine, later
   piglet_info.invalid_reason = LOC("$$$/PigletManager/12=Unknown reason")
   piglet_info.nonexistent = false


   -- Can't use LrFileUtils.exists() unless you're in a separate task... sigh.
   --    if not LrFileUtils.exists(piglet_info.file) then
   --       return false
   --    end

   local path = LrPathUtils.child(_PLUGIN.path, piglet_info.file)
   local handle = io.open(path, "r")
   if not handle then
      piglet_info.invalid_reason = LOC("$$$/PigletManager/13=Can't find or open piglet file^n(Did you remember to re-copy piglet files into the host plugin folder after a plugin upgrade?)")
      piglet_info.nonexistent = true
      return false
   else
      handle:close()
   end


   local okay, piglet = LrTasks.pcall(require, piglet_info.file)

   if not okay then
      local error_message = piglet
      runtime_error(piglet_info, nil, LOC("$$$/PigletManager/14=file load"), error_message)
      return false
   end

   if type(piglet) ~= 'table' then
      piglet_info.invalid_reason = LOC("$$$/PigletManager/15=Piglet contents seem to be corrupt")
      return false
   end

   if type(piglet.piglet_title) ~= 'string' then
      piglet_info.invalid_reason = LOC("$$$/PigletManager/16=Piglet file missing ^[piglet_title^] configuration")
      return false
   end

   piglet_info.piglet_title   = piglet.piglet_title
   piglet_info.piglet_desc    = piglet.piglet_desc
   piglet_info.piglet_url     = piglet.piglet_url
   piglet_info.piglet_version = piglet.piglet_version
   piglet_info.invalid = nil
   piglet_info.invalid_reason = nil

   if basic_info_only then
      return true
   end

   return piglet
end


local function get_piglet_file_list()
   local dest = LrPathUtils.child(_PLUGIN.path, "piglet.list")
   local cmd

   LrFileUtils.delete(dest)

   if WIN_ENV then

      local Qglob = '"'..LrPathUtils.child(_PLUGIN.path, "*.piglet")    ..'"'
      local Qdest = '"'..LrPathUtils.child(_PLUGIN.path, "piglet.list") ..'"'

      cmd = "dir /b " .. Qglob .. " > " .. Qdest .. ""
      local cmd_2run = '"' .. cmd .. '"'  -- quote everything, because Windows is stupid

      LrTasks.execute(cmd_2run)

   else

      local Qglob = LrPathUtils.child("'".._PLUGIN.path.."'", "*.piglet")
      local Qdest = LrPathUtils.child("'".._PLUGIN.path.."'", "piglet.list")

      cmd = 'ls -1 ' .. Qglob .. ' > ' .. Qdest .. ''
      LrTasks.execute(cmd)

   end


   -- want to check to see whether the file is actually there, and the only reasonable way to do that
   -- is by opening it.
   local handle = io.open(dest, "r")
   if not handle then
      LrDialogs.message(LOC("$$$/PigletManager/17=Couldn't create list-of-piglets file ^[^1^]", dest),
                        LOC("$$$/PigletManager/18=Attempted command:^n^1", cmd),
                        "critical")
      return {}
   end
   handle:close()    

   local files = {}
   for line in io.lines(dest) do
      table.insert(files, line)
   end
   return files
end



local MY_PIGLETS = { }
local piglets_loaded = false

local function sort_piglets(state)

   --
   -- Older versions of this package didn't have a "priority" field, so
   -- we've got to ensure that populate is populated.
   --
   for i, piglet_info in ipairs(state.piglet_info_list) do
      if not piglet_info.priority then
         piglet_info.priority = i
      end
   end

   table.sort(state.piglet_info_list,
              function(A, B)
                 if A.priority < B.priority then
                    return true
                 elseif A.priority > B.priority then
                    return false
                 else
                    return A.piglet_title < B.piglet_title
                 end
              end)

   table.sort(MY_PIGLETS,
              function(A, B)
                 if A._INFO.priority < B._INFO.priority then
                    return true
                 elseif A._INFO.priority > B._INFO.priority then
                    return false
                 else
                    return A._INFO.piglet_title < B._INFO.piglet_title
                 end
              end)

end

local function ensure_piglets_are_loaded(state) -- state is optional, and will be filled in if not provided

   if not piglets_loaded then

      piglets_loaded = true
      MY_PIGLETS = { }

      if not state then
         state = LrPrefs.prefsForPlugin(_PLUGIN.id)
      end

      --
      -- 
      -- This can be called very early in the plugin load sequence, even before
      -- state.piglet_info_list is initialized the very first time this is used in a
      -- new plugin, so we have to ensure that piglet_info_list exists before we
      -- access it.
      -- 
      --
      if state.piglet_info_list then

         for i, piglet_info in ipairs(state.piglet_info_list) do
            local piglet = read_piglet(state, piglet_info, false)
            if piglet_info.enable and piglet then
               -- 'piglet' and 'piglet_info' are kept separate because 'piglet_info' is saved across invocations, and we don't
               -- want to save the whole plugin there.
               piglet._INFO = piglet_info
               table.insert(MY_PIGLETS, piglet)

               --
               -- give some default priority if it's not already set
               -- (e.g. left over in the state from when a version of this plugin was used
               -- that didn't support priorites)
               --
               if not piglet_info.priority then
                  piglet_info.priority = i
               end
            end
         end

         sort_piglets(state)

      end
   end

end


local function update_all_piglets_info(state)

   ensure_piglets_are_loaded(state)

   local paths = get_piglet_file_list()

   if paths then

      for i, path in ipairs(paths) do
         
         local file = LrPathUtils.leafName(path)

         if file:match('%.piglet$') then
            local is_new = true
            for j, old_info in ipairs(state.piglet_info_list) do
               if file == old_info.file then
                  is_new = false
                  break
               end
            end
            if is_new then
               piglet_info = {
                  FOO = 'BAR',
                  file    = file,
                  invalid = nil,
                  enable  = false,
                  invalid_reason = nil,
                  nonexistant    = false,
                  priority       = 1 + #state.piglet_info_list,
               }
               read_piglet(state, piglet_info, true)
               table.insert(state.piglet_info_list, piglet_info)
            end
         end
      end
   end

end

local function show_manage_dialog(context, state)

   local v = LrView.osFactory()
   local props = LrBinding.makePropertyTable(context)

   update_all_piglets_info(state)
   sort_piglets(state)

   local COUNT = #state.piglet_info_list

   local DIALOG

   if COUNT == 0 then

      DIALOG = v:view {
         margin = 3,
         margin_bottom = 3,
         spacing = 15,
         v:static_text { title = LOC("$$$/PigletManager/19=No ^[*.piglet^] files found in this plugin's folder:") },
         v:static_text { title = "    " .. _PLUGIN.path,
                         text_color = LrColor(0, 0, 0.4) },

         v:static_text { title = LOC("$$$/PigletManager/20=Move piglet files into the folder for each plugin that you'd like to use them in.") },

         v:push_button {
            title = LOC("$$$/PigletManager/21=Get Piglets (Visit Website)"),
            action = function(button)
                        LrHttp.openUrlInBrowser("http://regex.info/blog/photo-tech/piglets/")
                     end,
            tooltip = LOC("$$$/PigletManager/22=Click to visit Jeffrey's piglet page"),
         },
      }


   else


      local priority_items = {}
      for i, piglet_info in ipairs(state.piglet_info_list) do
         table.insert(priority_items, { title = tostring(i), value = tostring(i) })
      end

      local ITEM_LIST = {
         place = 'vertical',
         spacing = 10,
      }

      if COUNT > 1 then
         table.insert(ITEM_LIST,
                      v:view {
                         margin_bottom = 1,
                         v:row{
                            v:static_text { title = LOC("$$$/PigletManager/32=Enabled"),      font = '<system/bold>', width = LrView.share("width_A") },
                            v:static_text { title = LOC("$$$/PigletManager/33=Priority"),     font = '<system/bold>', width = LrView.share("width_B") },
                            v:static_text { title = LOC("$$$/PigletManager/34=Description"),  font = '<system/bold>', width = LrView.share("width_C") },
                         },
                      })
      else
         table.insert(ITEM_LIST, v:row{
                         v:static_text { title = LOC("$$$/PigletManager/32=Enabled"),      font = '<system/bold>', width = LrView.share("width_A") },
                         v:static_text { title = LOC("$$$/PigletManager/34=Description"),  font = '<system/bold>', width = LrView.share("width_C") },
                      })
      end


      for i, piglet_info in ipairs(state.piglet_info_list) do

         local description

         if not piglet_info.piglet_version then
            if piglet_info.piglet_title then
               description = piglet_info.piglet_title .. " (" .. piglet_info.file .. ")"
            else
               description = piglet_info.file
            end
         else
            local version = LOC("$$$/PigletManager/36=version ^1", piglet_info.piglet_version)
            if piglet_info.piglet_title then
               description = piglet_info.piglet_title .. " (" .. piglet_info.file ..'; ' .. version .. ')'
            else
               description = piglet_info.file .. " (" .. version .. ")"
            end
         end


         local THIS_DESC = {
            place = 'vertical',
            width = LrView.share("width_C"),

            v:static_text {
               title = description,
               text_color = titleColor
            },
         }


         if piglet_info.piglet_desc then
            table.insert(THIS_DESC,
                         v:static_text {
                            title = piglet_info.piglet_desc,
                            text_color = LrColor(0.4, 0.4, 0.4),
                         })
         end

         -- ensure that the priorities stay in the range of the priority_items table
         if piglet_info.priority > COUNT then
            piglet_info.priority = COUNT
         end

         local titleColor = nil

         if piglet_info.invalid then

            titleColor = LrColor(0.4, 0.4, 0.4) -- a "disabled" looking color

            local error_guts = {
               v:static_text {
                  title = piglet_info.invalid_reason,
                  text_color = LrColor(1,0,0)
               }
            }

            if piglet_info.nonexistent then
               local key = 'delete_' .. tostring(i)
               props[key] = false
               table.insert(error_guts, v:checkbox {
                               title = LOC("$$$/PigletManager/23=check to remove this piglet from this list"),
                               value = LrView.bind(key),
                            })
            end

            if piglet_info.piglet_desc then
               table.insert(THIS_DESC, v:view(error_guts))
            end
         end


         local visit_homepage
         if not piglet_info.piglet_url then
            visit_homepage = v:spacer { width = 1 } -- dummy spacer
         else
            visit_homepage = v:push_button {
               size = 'mini',
               title = LOC("$$$/PigletManager/24=visit homepage"),
               action = function(button)
                           LrHttp.openUrlInBrowser(piglet_info.piglet_url)
                        end,
            }
         end

         local PRIORITY
         
         if COUNT == 1 then
            PRIORITY = v:spacer { width = 1 } -- dummy spacer
         else
            PRIORITY = v:popup_menu {
               items = priority_items,
               enabled = not piglet_info.invalid,
               width_in_chars = 2,
               width = LrView.share("width_B"),
               value = LrView.bind {
                  key = 'priority_' .. tostring(i),
                  transform = function(newValue, fromTable)
                                 if not fromTable then
                                    piglet_info.priority = tonumber(newValue)
                                 end
                                 return tostring(piglet_info.priority)
                              end
               }
            }
         end

         table.insert(ITEM_LIST,
                      v:row {
                         v:checkbox {
                            title = "",
                            enabled = not piglet_info.invalid,
                            width = LrView.share("width_A"),
                            value = LrView.bind {
                               key = 'checkbox_' .. tostring(i),
                               transform = function(newValue, fromTable)
                                              if not fromTable then
                                                 piglet_info.enable = newValue
                                              end
                                              return piglet_info.enable
                                           end
                            },
                         },

                         PRIORITY,
                         v:view(THIS_DESC),
                         visit_homepage,
                      }
                   )
      end
      
      if COUNT > 1 then
         table.insert(ITEM_LIST, v:static_text { title = LOC("$$$/PigletManager/35=Priority 1 piglets execute before priority 2 piglets, etc.") })
      end

      DIALOG = v:view {
         place = 'vertical',
         spacing = 10,
         bind_to_object = props,

         v:static_text { title = LOC("$$$/PigletManager/25=Piglets available to this host plugin:") },

         v:group_box {
            title = "",
            show_title = false,
            spacing = 10,

            v:view(ITEM_LIST),
         },

         v:view {
            margin_top = 10,
            margin_bottom = 20,

            v:separator {
               fill_horizontal = 1
            },
            v:static_text {
               text_color = LrColor(0.9,0,0),
               title = LOC("$$$/PigletManager/26=Changes take effect next time you RESTART Lightroom"),
            },
            v:static_text {
               title = LOC("$$$/PigletManager/27=Yes, it's lame, but that's the way it is. Sorry."),
               text_color = LrColor(0.4, 0.4, 0.4)
            },
         },


         v:push_button {
            title = LOC("$$$/PigletManager/21=Get Piglets (Visit Website)"),
            action = function(button)
                        LrHttp.openUrlInBrowser("http://regex.info/blog/photo-tech/piglets/")
                     end,
            tooltip = LOC("$$$/PigletManager/22=Click to visit Jeffrey's piglet page"),
         },
      }

   end

   local UserReply = LrDialogs.presentModalDialog {
      save_frame = 'piglet_manager_frame',
      title = LOC("$$$/PigletManager/28=Piglet Manager"),
      resizable = true,
      actionVerb = LOC("$$$/PigletManager/29=Done"),
      cancelVerb = "< exclude >", -- this magic, unpublished setting removes the cancel button, and may not work in the future
      contents = DIALOG,
   }

   --
   -- Check in reverse order so that we can simply table.remove() them without worrying about shifting elements
   --
   for i = COUNT, 1, -1 do
      if props['delete_' .. tostring(i)] then
         table.remove(state.piglet_info_list, i)
      end
   end

end

function Piglets.button(v, state, ph)

   return v:push_button {
      place_horizontal = ph,
      title = LOC("$$$/PigletManager/30=Manage plugin additions (^]piglets^])"),
      action = function(button)
                  LrFunctionContext.postAsyncTaskWithContext("piglet manager",
                                                             function(context)
                                                                local okay, msg = LrTasks.pcall(show_manage_dialog, context, state)
                                                                if not okay then
                                                                   LrDialogs.message(LOC("$$$/101=INTERNAL PLUGIN ERROR"),
                                                                                     msg,
                                                                                     "critical")
                                                                end
                                                             end)
               end
   }

end

function Piglets.exportPresetFields()

   ensure_piglets_are_loaded()

   local preset_fields = {
      {   key = 'piglet_info_list',   default = { }   }
   }

   for i, piglet in ipairs(MY_PIGLETS) do
      if type(piglet.exportPresetFields) == 'table' and #piglet.exportPresetFields >= 1 then
         concatenate(preset_fields, piglet.exportPresetFields)
      end
   end
   return preset_fields
end



--
-- We want to allow any file format allowed by the host plugin and any piglet,
-- so if any of them don't restrict the formats, all formats are allowed.
--
-- The formats allowed by the host plugin are passed in a table as the first argument.
--
function Piglets.allowFileFormats(allowed)

   ensure_piglets_are_loaded()

   if #MY_PIGLETS == 0 then
      -- if there are no piglets, than just return what the host plugin allows.
      return allowed
   end

   --
   -- If the host plugin wants to allow all formats, it should just skip
   -- this entirely and leave alllowFileFormats nil, but just in case that
   -- nil is passed here, we should handle it.
   --
   if allowed == nil then
      -- if the host plugin allows all file formats
      return allowed
   end


   -- where we'll accmuluate the formats allowed by the piglets
   local vals = { }

   --
   -- Go through the piglets, adding each piglet's allowed formats to the
   -- running collection of formats.
   --
   for i, piglet in ipairs(MY_PIGLETS) do
      if piglet.allowFileFormats == nil then
         -- a piglet allows all formats, so we tell the host plugin to do the same
         return nil
      end

      if type(piglet.allowFileFormats) == table then
         for i, val in ipairs(piglet.allowFileFormats) do
            vals[string.upper(val)] = true
         end
      end
   end

   --
   -- We be sure to include in the running collection of formats those allowed
   -- by the host plugin.
   --
   if type(allowed) == table then
      for i, val in ipairs(allowed) do
         vals[string.upper(val)] = true
      end
   end


   local return_val = { }

   -- Turn 'vals' into an array of values
   for k, v in pairs(vals) do
      table.insert(return_val, k)
   end

   if #return_val == 0 then
      return nil
   else
      return return_val
   end
end


function Piglets.startDialog(state)

   ensure_piglets_are_loaded(state)

   for i, piglet in ipairs(MY_PIGLETS) do
      if not piglet._INFO.invalid then
         if type(piglet.startDialog) == 'function' then

            local okay, error_msg = LrTasks.pcall(piglet.startDialog, state)
            if not okay then
               runtime_error(piglet._INFO, piglet, 'startDialog()', error_msg)
            end
         end
      end
   end

end


function Piglets.endDialog(state, why)

   ensure_piglets_are_loaded(state)

   for i, piglet in ipairs(MY_PIGLETS) do
      if not piglet._INFO.invalid then
         if type(piglet.endDialog) == 'function' then

            local okay, error_msg = LrTasks.pcall(piglet.endDialog, state, why)
            if not okay then
               runtime_error(piglet._INFO, piglet, 'endDialog()', error_msg)
            end

         end
      end
   end

end


function Piglets.appendTopSections(Sections, state, v)

   ensure_piglets_are_loaded(state)

   for i, piglet in ipairs(MY_PIGLETS) do
      if not piglet._INFO.invalid then
         if type(piglet.sectionsForTopOfDialog) == 'function' then

            local okay, retval = LrTasks.pcall(piglet.sectionsForTopOfDialog, v, state)
            if not okay then
               runtime_error(piglet._INFO, piglet, 'sectionsForTopOfDialog()', retval)
            end
            local new = retval

            if type(new) == 'table' and #new >= 1 then
               for i, section in ipairs(new) do
                  section.title = LOC("$$$/PigletManager/31=PIGLET") ..  ": " .. tostring(section.title)
                  table.insert(Sections, section)
               end
            end
         end
      end
   end
   return Sections -- Sections is updated in place, but we return it anyway, in case it might be convenient

end


function Piglets.appendBottomSections(Sections, state, v)

   ensure_piglets_are_loaded(state)

   for i, piglet in ipairs(MY_PIGLETS) do
      if not piglet._INFO.invalid then
         if type(piglet.sectionsForBottomOfDialog) == 'function' then

            local okay, retval = LrTasks.pcall(piglet.sectionsForBottomOfDialog, v, state)
            if not okay then

               runtime_error(piglet._INFO, piglet, 'sectionsForBottomOfDialog()', retval)

            end
            local new = retval

            if type(new) == 'table' and #new >= 1 then
               for i, section in ipairs(new) do
                  section.title = LOC("$$$/PigletManager/31=PIGLET") ..": " .. tostring(section.title)
                  table.insert(Sections, section)
               end
            end
         end
      end
   end
   return Sections -- Sections is updated in place, but we return it anyway, in case it might be convenient

end

function Piglets.set_CanExport_Observer(state, func)

   ensure_piglets_are_loaded(state)

   for i, piglet in ipairs(MY_PIGLETS) do
      if not piglet._INFO.invalid then
         if piglet.canExportKey then
            state:addObserver(piglet.canExportKey, func)
         end
      end
   end

end



function Piglets.canExport(state)

   ensure_piglets_are_loaded(state)

   for i, piglet in ipairs(MY_PIGLETS) do
      if not piglet._INFO.invalid then
         if piglet.canExportKey and not state[piglet.canExportKey] then
            return false
         end
      end
   end

   return true
end



do
   local function_to_mirror = {
      'priorToRender',
      'startOfOneRender',
      'endOfOneRender',
      'exportWasCanceled',
      'exportComplete',
   }

   local function do_func(which, INFO)
      for i, piglet in ipairs(MY_PIGLETS) do

         -- make a clean LOCAL table for the piglet to use
         if which == 'priorToRender' then
            piglet.LOCAL = {}
         end

         if type(piglet[which]) == 'function' then

            INFO.LOCAL = piglet.LOCAL
            local okay, retval = LrTasks.pcall(piglet[which], INFO)
            if not okay then
               runtime_error(piglet._INFO, piglet, which .. "()", retval)
               return false
            elseif not retval then
               -- abort entire export
               return false
            elseif retval == 'skip this image' then
               return retval
            end
         end
      end

      return true
   end

   for i, func in ipairs(function_to_mirror) do
      Piglets[func] = function(INFO)
                         return do_func(func, INFO)
                      end
   end
end

return Piglets
-- [LOC PREFIX: PigletManager]
